/**
 * Copyright 2013,2015 IBM Corp.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

module.exports = function (RED) {
  'use strict';
  var request = require('request');
  var ffmpeg = require('easy-ffmpeg');
  var url = require('url');
  var temp = require('temp').track();
  var fs = require('fs');
  var path = require('path');

  var timeouts = [];

  // Utility function to perform a URL validation check. Copied from speech_to_text.
  function urlCheck(str) {
    var parsed = url.parse(str);

    return (!!parsed.hostname && !!parsed.protocol && str.indexOf(' ') < 0);
  }

  function Node (config) {
    RED.nodes.createNode(this, config);
    var node = this;

    this.on('input', function (msg) {
      // Cancel any waiting timeouts
      for (var i = 0; i < timeouts.length; ++i) {
        clearTimeout(timeouts[i]);
      }
      timeouts = [];
      node.status({});

      var message;

      // Return error and clean up
      function nodeError(nodeText, debugMsg) {
        if (nodeText) {
          node.status({fill:'red', shape:'ring', text: nodeText});
        }

        message = debugMsg;
        temp.cleanup();
        node.error(message, msg);
      }

      function nodeSend(silences) {
        msg.silences = silences;
        temp.cleanup();
        node.status({});
        node.send(msg);
      }

      // This section is for var functions that will be called, in context with
      // msg, when the input stream has been received.

      // Function that is syncing up the asynchronous nature of the stream
      // so that the full file can be sent to the API. Copied and edited from speech_to_text.
      var stream_buffer = function(file, contents, cb) {
        node.status({fill:'blue', shape:'dot',
          text:'reading buffer'});
        fs.writeFile(file, contents, function (err) {
          cb(err);
        });
      };

      // Function that is syncing up the asynchronous nature of the stream
      // so that the full file can be sent to the API. Copied and edited from speech_to_text.
      var stream_url = function(file, url, cb) {
        var wstream = fs.createWriteStream(file);
        node.status({fill:'blue', shape:'dot',
          text:'downloading file'});

        wstream
          .on('error', function(err) {
            cb(err);
          })
          .on('finish', function () {
            cb();
          });

        request(url).pipe(wstream);
      };

      // Get segment times of approximate silence
      var getSegments = function(path, cb) {
        var silences = [];
        var lastStart, lastEnd;

        // Silence start and end points are outputted on stderr. We read each of these lines to parse times and generate average timestamps
        var silenceDetectStdErr = function(stderrLine) {
          var startStr = 'silence_start: ';
          var endStr = 'silence_end: ';
          var durStr = ' | silence_duration:';
          var iStart = parseInt(stderrLine.indexOf(startStr), 10);
          var iEnd = stderrLine.indexOf(endStr);

          if (iStart > -1) {
            lastStart = parseFloat(stderrLine.slice(iStart + startStr.length, stderrLine.length), 10);
          } else if (iEnd > -1) {
            var iDur = parseInt(stderrLine.indexOf(durStr), 10);
            lastEnd = parseFloat(stderrLine.slice(iEnd + endStr.length, iDur), 10);
            silences.push([lastStart, lastEnd]);
          }
        };

        var silenceDetectError = function() {
          nodeError('ffmpeg conversion failed', 'FFmpeg failed to perform the conversion');
        };

        // Once ffmpeg segments are read, call callback
        var silenceDetectEnd = function() {
          cb(silences);
        };

        var filters = 'silencedetect=n=' + config.noise + ':d=' + config.duration;

        ffmpeg(path)
          .format(null)
          .audioFilters(filters)
          .on('stderr', silenceDetectStdErr)
          .on('error', silenceDetectError)
          .on('end', silenceDetectEnd)
          .output(temp.path({suffix: '.' + config.format}))
          .run();
      };

      // Performs action to get silence segments
      function performAction(pathToFile) {
        getSegments(pathToFile, function(silences) {
          nodeSend(silences);
        });
      }

      // Check payload
      if (!msg.payload) {
        return nodeError(null, 'Missing property: msg.payload');
      }

      // Check payload
      if (!(msg.payload instanceof Buffer) && typeof msg.payload === 'string' && !urlCheck(msg.payload)) {
        return nodeError(null, 'Invalid URL.');
      }

      if (msg.payload instanceof Buffer) {
        temp.open({suffix: '.buffer'}, function (err, info) {
          if (err) {
            return nodeError('unable to open buffer', 'Node has been unable to open the buffer');
          }

          // Stream buffer into temp file
          stream_buffer(info.path, msg.payload, function (err) {
            if (err) {
              return nodeError('unable to open buffer', 'Node has been unable to open the buffer');
            }

            performAction(info.path);
          });
        });
      } else {
        var ext = path.extname(msg.payload);
        temp.open({suffix: ext}, function(err, info) {
          if (err) {
            return nodeError('unable to open url stream', 'Node has been unable to open the url stream');
          }

          stream_url(info.path, msg.payload, function (err) {
            if (err) {
              return nodeError('url stream not recognised', 'Node did not recognise the url stream');
            }

            performAction(info.path);
          });
        });
      }
    });
  }
  RED.nodes.registerType('ffmpeg-silence-detect', Node);
};
